/*******************************************************************************
 * Copyright (c) 2000, 2007 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
module dwt.dnd;
 
import dwt.internal.carbon.OS;

/**
 * The class <code>HTMLTransfer</code> provides a platform specific mechanism 
 * for converting text in HTML format represented as a java <code>String</code> 
 * to a platform specific representation of the data and vice versa.  See 
 * <code>Transfer</code> for additional information.
 * 
 * <p>An example of a java <code>String</code> containing HTML text is shown 
 * below:</p>
 * 
 * <code><pre>
 *     String htmlData = "<p>This is a paragraph of text.</p>";
 * </code></pre>
 */
public class HTMLTransfer : ByteArrayTransfer {

    static HTMLTransfer _instance = new HTMLTransfer();
    static final String HTML = "HTML"; //$NON-NLS-1$
    static final int HTMLID = registerType(HTML);

this() {}

/**
 * Returns the singleton instance of the HTMLTransfer class.
 *
 * @return the singleton instance of the HTMLTransfer class
 */
public static HTMLTransfer getInstance () {
    return _instance;
}

/**
 * This implementation of <code>javaToNative</code> converts HTML-formatted text
 * represented by a java <code>String</code> to a platform specific representation.
 * For additional information see <code>Transfer#javaToNative</code>.
 * 
 * @param object a java <code>String</code> containing HTML text
 * @param transferData an empty <code>TransferData</code> object; this
 *  object will be filled in on return with the platform specific format of the data
 */
public void javaToNative (Object object, TransferData transferData){
    if (!checkHTML(object) || !isSupportedType(transferData)) {
        DND.error(DND.ERROR_INVALID_DATA);
    }
    String String = (String)object;
    int count = String.length();
    char[] chars = new char[count];
    String.getChars(0, count, chars, 0);
    byte[] buffer = new byte[chars.length * 2];
    OS.memmove(buffer, chars, buffer.length);
    transferData.data = new byte[1][];
    transferData.data[0] = buffer;
    transferData.result = OS.noErr;
}

/**
 * This implementation of <code>nativeToJava</code> converts a platform specific 
 * representation of HTML text to a java <code>String</code>.
 * For additional information see <code>Transfer#nativeToJava</code>.
 * 
 * @param transferData the platform specific representation of the data to be 
 * been converted
 * @return a java <code>String</code> containing HTML text if the 
 * conversion was successful; otherwise null
 */
public Object nativeToJava(TransferData transferData){
    if (!isSupportedType(transferData) || transferData.data is null) return null;
    if (transferData.data.length is 0 || transferData.data[0].length is 0) return null;
    byte[] buffer = transferData.data[0];
    char[] chars = new char[(buffer.length + 1) / 2];
    OS.memmove(chars, buffer, buffer.length);
    return new String(chars);
}

protected int[] getTypeIds() {
    return new int[] {HTMLID};
}

protected String[] getTypeNames() {
    return new String[] {HTML};
}

bool checkHTML(Object object) {
    return (object !is null && object instanceof String && ((String)object).length() > 0);
}

protected bool validate(Object object) {
    return checkHTML(object);
}
}
