/*******************************************************************************
 * Copyright (c) 2000, 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
module dwt.dnd;
 
import dwt.internal.carbon.OS;
import dwt.internal.carbon.CFRange;

/**
 * The class <code>RTFTransfer</code> provides a platform specific mechanism 
 * for converting text in RTF format represented as a java <code>String</code> 
 * to a platform specific representation of the data and vice versa.  See 
 * <code>Transfer</code> for additional information.
 * 
 * <p>An example of a java <code>String</code> containing RTF text is shown 
 * below:</p>
 * 
 * <code><pre>
 *     String rtfData = "{\\rtf1{\\colortbl;\\red255\\green0\\blue0;}\\uc1\\b\\i Hello World}";
 * </code></pre>
 */
public class RTFTransfer : ByteArrayTransfer {

    static RTFTransfer _instance = new RTFTransfer();
    static final String RTF = "RTF "; //$NON-NLS-1$
    static final int RTFID = registerType(RTF);

this() {}

/**
 * Returns the singleton instance of the RTFTransfer class.
 *
 * @return the singleton instance of the RTFTransfer class
 */
public static RTFTransfer getInstance () {
    return _instance;
}

/**
 * This implementation of <code>javaToNative</code> converts RTF-formatted text
 * represented by a java <code>String</code> to a platform specific representation.
 * For additional information see <code>Transfer#javaToNative</code>.
 * 
 * @param object a java <code>String</code> containing RTF text
 * @param transferData an empty <code>TransferData</code> object; this
 *  object will be filled in on return with the platform specific format of the data
 */
public void javaToNative (Object object, TransferData transferData){
    if (!checkRTF(object) || !isSupportedType(transferData)) {
        DND.error(DND.ERROR_INVALID_DATA);
    }
    transferData.result = -1;
    String String = (String)object;
    int count = String.length();
    char[] chars = new char[count];
    String.getChars(0, count, chars, 0);
    int cfString = OS.CFStringCreateWithCharacters(OS.kCFAllocatorDefault, chars, count);
    if (cfString is 0) return;
    try {
        CFRange range = new CFRange();
        range.length = chars.length;
        int encoding = OS.CFStringGetSystemEncoding();
        int[] size = new int[1];
        int numChars = OS.CFStringGetBytes(cfString, range, encoding, (byte)'?', true, null, 0, size);
        if (numChars is 0 || size[0] is 0) return;
        byte[] buffer = new byte[size[0]];
        numChars = OS.CFStringGetBytes(cfString, range, encoding, (byte)'?', true, buffer, size [0], size);
        if (numChars is 0) return;
        transferData.data = new byte[1][];
        transferData.data[0] = buffer;
        transferData.result = 0;
    } finally {
        OS.CFRelease(cfString);
    }
}

/**
 * This implementation of <code>nativeToJava</code> converts a platform specific 
 * representation of RTF text to a java <code>String</code>.
 * For additional information see <code>Transfer#nativeToJava</code>.
 * 
 * @param transferData the platform specific representation of the data to be 
 * been converted
 * @return a java <code>String</code> containing RTF text if the 
 * conversion was successful; otherwise null
 */
public Object nativeToJava(TransferData transferData){
    if (!isSupportedType(transferData) || transferData.data is null) return null;
    if (transferData.data.length is 0 || transferData.data[0].length is 0) return null;
    byte[] buffer = transferData.data[0];
    int encoding = OS.CFStringGetSystemEncoding();
    int cfString = OS.CFStringCreateWithBytes(OS.kCFAllocatorDefault, buffer, buffer.length, encoding, true);
    if (cfString is 0) return null;
    try {
        int length = OS.CFStringGetLength(cfString);
        if (length is 0) return null;
        char[] chars = new char[length];
        CFRange range = new CFRange();
        range.length = length;
        OS.CFStringGetCharacters(cfString, range, chars);
        return new String(chars);
    } finally {
        OS.CFRelease(cfString);
    }
}

protected int[] getTypeIds() {
    return new int[] {RTFID};
}

protected String[] getTypeNames() {
    return new String[] {RTF};
}

bool checkRTF(Object object) {
    return (object !is null && object instanceof String && ((String)object).length() > 0);
}

protected bool validate(Object object) {
    return checkRTF(object);
}
}
