/*******************************************************************************
 * Copyright (c) 2000, 2007 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
module dwt.graphics.Font;

import dwt.dwthelper.utils;


import dwt.DWT;
import dwt.DWTError;
import dwt.DWTException;
import dwt.internal.cocoa.NSFont;
import dwt.internal.cocoa.NSString;

/**
 * Instances of this class manage operating system resources that
 * define how text looks when it is displayed. Fonts may be constructed
 * by providing a device and either name, size and style information
 * or a <code>FontData</code> object which encapsulates this data.
 * <p>
 * Application code must explicitly invoke the <code>Font.dispose()</code> 
 * method to release the operating system resources managed by each instance
 * when those instances are no longer required.
 * </p>
 *
 * @see FontData
 */
public final class Font : Resource {

    /**
     * the handle to the OS font resource
     * (Warning: This field is platform dependent)
     * <p>
     * <b>IMPORTANT:</b> This field is <em>not</em> part of the DWT
     * public API. It is marked public only so that it can be shared
     * within the packages provided by DWT. It is not available on all
     * platforms and should never be accessed from application code.
     * </p>
     */
    public NSFont handle;
    
this(Device device) {
    super(device);
}

/**  
 * Constructs a new font given a device and font data
 * which describes the desired font's appearance.
 * <p>
 * You must dispose the font when it is no longer required. 
 * </p>
 *
 * @param device the device to create the font on
 * @param fd the FontData that describes the desired font (must not be null)
 * 
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_NULL_ARGUMENT - if device is null and there is no current device</li>
 *    <li>ERROR_NULL_ARGUMENT - if the fd argument is null</li>
 * </ul>
 * @exception DWTError <ul>
 *    <li>ERROR_NO_HANDLES - if a font could not be created from the given font data</li>
 * </ul>
 */
public this(Device device, FontData fd) {
    super(device);
    if (fd is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    init(fd.getName(), fd.getHeightF(), fd.getStyle(), fd.nsName);
    init();
}

/**  
 * Constructs a new font given a device and an array
 * of font data which describes the desired font's
 * appearance.
 * <p>
 * You must dispose the font when it is no longer required. 
 * </p>
 *
 * @param device the device to create the font on
 * @param fds the array of FontData that describes the desired font (must not be null)
 * 
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_NULL_ARGUMENT - if device is null and there is no current device</li>
 *    <li>ERROR_NULL_ARGUMENT - if the fds argument is null</li>
 *    <li>ERROR_INVALID_ARGUMENT - if the length of fds is zero</li>
 *    <li>ERROR_NULL_ARGUMENT - if any fd in the array is null</li>
 * </ul>
 * @exception DWTError <ul>
 *    <li>ERROR_NO_HANDLES - if a font could not be created from the given font data</li>
 * </ul>
 * 
 * @since 2.1
 */
public this(Device device, FontData[] fds) {
    super(device);
    if (fds is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (fds.length is 0) DWT.error(DWT.ERROR_INVALID_ARGUMENT);
    for (int i=0; i<fds.length; i++) {
        if (fds[i] is null) DWT.error(DWT.ERROR_INVALID_ARGUMENT);
    }
    FontData fd = fds[0];
    init(fd.getName(), fd.getHeightF(), fd.getStyle(), fd.nsName);
    init();
}

/**  
 * Constructs a new font given a device, a font name,
 * the height of the desired font in points, and a font
 * style.
 * <p>
 * You must dispose the font when it is no longer required. 
 * </p>
 *
 * @param device the device to create the font on
 * @param name the name of the font (must not be null)
 * @param height the font height in points
 * @param style a bit or combination of NORMAL, BOLD, ITALIC
 * 
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_NULL_ARGUMENT - if device is null and there is no current device</li>
 *    <li>ERROR_NULL_ARGUMENT - if the name argument is null</li>
 *    <li>ERROR_INVALID_ARGUMENT - if the height is negative</li>
 * </ul>
 * @exception DWTError <ul>
 *    <li>ERROR_NO_HANDLES - if a font could not be created from the given arguments</li>
 * </ul>
 */
public this(Device device, String name, int height, int style) {
    super(device);
    init(name, height, style, null);
    init();
}

/*public*/ Font(Device device, String name, float height, int style) {
    super(device);
    init(name, height, style, null);
    init();
}

void destroy() {
    handle.release();
    handle = null;
}

/**
 * Compares the argument to the receiver, and returns true
 * if they represent the <em>same</em> object using a class
 * specific comparison.
 *
 * @param object the object to compare with this object
 * @return <code>true</code> if the object is the same as this object and <code>false</code> otherwise
 *
 * @see #hashCode
 */
public bool equals(Object object) {
    if (object is this) return true;
    if (!(object instanceof Font)) return false;
    Font font = (Font)object;
    return handle is font.handle;
}

/**
 * Returns an array of <code>FontData</code>s representing the receiver.
 * On Windows, only one FontData will be returned per font. On X however, 
 * a <code>Font</code> object <em>may</em> be composed of multiple X 
 * fonts. To support this case, we return an array of font data objects.
 *
 * @return an array of font data objects describing the receiver
 *
 * @exception DWTException <ul>
 *    <li>ERROR_GRAPHIC_DISPOSED - if the receiver has been disposed</li>
 * </ul>
 */
public FontData[] getFontData() {
    if (isDisposed()) DWT.error(DWT.ERROR_GRAPHIC_DISPOSED);
    NSString family = handle.familyName();
    char[] buffer1 = new char[family.length()];
    family.getCharacters_(buffer1);
    String name = new String(buffer1);
    NSString str = handle.fontName();
    char[] buffer = new char[str.length()];
    str.getCharacters_(buffer);
    String nsName = new String(buffer);
    int style = DWT.NORMAL;
    if (nsName.indexOf("Italic") !is -1) style |= DWT.ITALIC;
    if (nsName.indexOf("Bold") !is -1) style |= DWT.BOLD;
    FontData data = new FontData(name, handle.pointSize(), style);
    data.nsName = nsName;
    return new FontData[]{data};
}

/**  
 * Invokes platform specific functionality to allocate a new font.
 * <p>
 * <b>IMPORTANT:</b> This method is <em>not</em> part of the public
 * API for <code>Font</code>. It is marked public only so that it
 * can be shared within the packages provided by DWT. It is not
 * available on all platforms, and should never be called from
 * application code.
 * </p>
 *
 * @param device the device on which to allocate the color
 * @param handle the handle for the font
 * @param style the style for the font
 * @param size the size for the font
 * 
 * @private
 */
public static Font cocoa_new(Device device, NSFont handle) {
    Font font = new Font(device);
    font.handle = handle;
    return font;
}

/**
 * Returns an integer hash code for the receiver. Any two 
 * objects that return <code>true</code> when passed to 
 * <code>equals</code> must return the same value for this
 * method.
 *
 * @return the receiver's hash
 *
 * @see #equals
 */
public int hashCode() {
    return handle !is null ? handle.id : 0;
}

void init(String name, float height, int style, String nsName) {
    if (name is null) DWT.error(DWT.ERROR_NULL_ARGUMENT);
    if (height < 0) DWT.error(DWT.ERROR_INVALID_ARGUMENT);
    if (nsName !is null) {
        handle = NSFont.static_fontWithName_size_(NSString.stringWith(nsName), height);
    } else {
        nsName = name;
        if ((style & DWT.BOLD) !is 0) nsName += " Bold";
        if ((style & DWT.ITALIC) !is 0) nsName += " Italic";
        handle = NSFont.static_fontWithName_size_(NSString.stringWith(nsName), height);
        if (handle is null && (style & DWT.ITALIC) !is 0) {
            nsName = name;
            if ((style & DWT.BOLD) !is 0) nsName += " Bold";
            handle = NSFont.static_fontWithName_size_(NSString.stringWith(nsName), height);
        }
        if (handle is null && (style & DWT.BOLD) !is 0) {
            nsName = name;
            handle = NSFont.static_fontWithName_size_(NSString.stringWith(nsName), height);
        }
    }
    if (handle is null) {
        handle = device.systemFont.handle;
    }
    handle.retain();
}

/**
 * Returns <code>true</code> if the font has been disposed,
 * and <code>false</code> otherwise.
 * <p>
 * This method gets the dispose state for the font.
 * When a font has been disposed, it is an error to
 * invoke any other method using the font.
 *
 * @return <code>true</code> when the font is disposed and <code>false</code> otherwise
 */
public bool isDisposed() {
    return handle is null;
}

/**
 * Returns a string containing a concise, human-readable
 * description of the receiver.
 *
 * @return a string representation of the receiver
 */
public String toString () {
    if (isDisposed()) return "Font {*DISPOSED*}";
    return "Font {" + handle + "}";
}

}
