/*******************************************************************************
 * Copyright (c) 2000, 2007 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
module dwt.widgets.MessageBox;

import dwt.dwthelper.utils;


import dwt.DWT;
import dwt.DWTException;
import dwt.internal.cocoa.NSAlert;
import dwt.internal.cocoa.NSString;
import dwt.internal.cocoa.OS;

/**
 * Instances of this class are used to inform or warn the user.
 * <dl>
 * <dt><b>Styles:</b></dt>
 * <dd>ICON_ERROR, ICON_INFORMATION, ICON_QUESTION, ICON_WARNING, ICON_WORKING</dd>
 * <dd>OK, OK | CANCEL</dd>
 * <dd>YES | NO, YES | NO | CANCEL</dd>
 * <dd>RETRY | CANCEL</dd>
 * <dd>ABORT | RETRY | IGNORE</dd>
 * <dt><b>Events:</b></dt>
 * <dd>(none)</dd>
 * </dl>
 * <p>
 * Note: Only one of the styles ICON_ERROR, ICON_INFORMATION, ICON_QUESTION,
 * ICON_WARNING and ICON_WORKING may be specified.
 * </p><p>
 * IMPORTANT: This class is intended to be subclassed <em>only</em>
 * within the DWT implementation.
 * </p>
 */
public  class MessageBox : Dialog {
    String message = "";
    

/**
 * Constructs a new instance of this class given only its parent.
 *
 * @param parent a shell which will be the parent of the new instance
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_NULL_ARGUMENT - if the parent is null</li>
 * </ul>
 * @exception DWTException <ul>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the parent</li>
 *    <li>ERROR_INVALID_SUBCLASS - if this class is not an allowed subclass</li>
 * </ul>
 */
public this (Shell parent) {
    this (parent, DWT.OK | DWT.ICON_INFORMATION | DWT.APPLICATION_MODAL);
}

/**
 * Constructs a new instance of this class given its parent
 * and a style value describing its behavior and appearance.
 * <p>
 * The style value is either one of the style constants defined in
 * class <code>DWT</code> which is applicable to instances of this
 * class, or must be built by <em>bitwise OR</em>'ing together 
 * (that is, using the <code>int</code> "|" operator) two or more
 * of those <code>DWT</code> style constants. The class description
 * lists the style constants that are applicable to the class.
 * Style bits are also inherited from superclasses.
 *
 * @param parent a shell which will be the parent of the new instance
 * @param style the style of dialog to construct
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_NULL_ARGUMENT - if the parent is null</li>
 * </ul>
 * @exception DWTException <ul>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the parent</li>
 *    <li>ERROR_INVALID_SUBCLASS - if this class is not an allowed subclass</li>
 * </ul>
 */
public this (Shell parent, int style) {
    super (parent, checkStyle (style));
    checkSubclass ();
}

static int checkStyle (int style) {
    if ((style & (DWT.PRIMARY_MODAL | DWT.APPLICATION_MODAL | DWT.SYSTEM_MODAL)) is 0) style |= DWT.APPLICATION_MODAL;
    int mask = (DWT.YES | DWT.NO | DWT.OK | DWT.CANCEL | DWT.ABORT | DWT.RETRY | DWT.IGNORE);
    int bits = style & mask;
    if (bits is DWT.OK || bits is DWT.CANCEL || bits is (DWT.OK | DWT.CANCEL)) return style;
    if (bits is DWT.YES || bits is DWT.NO || bits is (DWT.YES | DWT.NO) || bits is (DWT.YES | DWT.NO | DWT.CANCEL)) return style;
    if (bits is (DWT.RETRY | DWT.CANCEL) || bits is (DWT.ABORT | DWT.RETRY | DWT.IGNORE)) return style;
    style = (style & ~mask) | DWT.OK;
    return style;
}

/**
 * Returns the dialog's message, or an empty string if it does not have one.
 * The message is a description of the purpose for which the dialog was opened.
 * This message will be visible in the dialog while it is open.
 *
 * @return the message
 */
public String getMessage () {
    return message;
}

/**
 * Makes the dialog visible and brings it to the front
 * of the display.
 *
 * @return the ID of the button that was selected to dismiss the
 *         message box (e.g. DWT.OK, DWT.CANCEL, etc.)
 *
 * @exception DWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the dialog has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the dialog</li>
 * </ul>
 */
public int open () {
    int alertType = OS.NSInformationalAlertStyle;
    if ((style & DWT.ICON_ERROR) !is 0) alertType = OS.NSCriticalAlertStyle;
    if ((style & DWT.ICON_INFORMATION) !is 0) alertType = OS.NSInformationalAlertStyle;
    if ((style & DWT.ICON_QUESTION) !is 0) alertType = OS.NSInformationalAlertStyle;
    if ((style & DWT.ICON_WARNING) !is 0) alertType = OS.NSWarningAlertStyle;
    if ((style & DWT.ICON_WORKING) !is 0) alertType = OS.NSInformationalAlertStyle;
    NSString defaultButton = null, alternateButton = null, otherButton = null;
    int mask = (DWT.YES | DWT.NO | DWT.OK | DWT.CANCEL | DWT.ABORT | DWT.RETRY | DWT.IGNORE);
    int bits = style & mask;
    switch (bits) {
        case DWT.OK:
            break;
        case DWT.CANCEL:
            defaultButton = NSString.stringWith(DWT.getMessage("SWT_Cancel"));
            break;
        case DWT.OK | DWT.CANCEL:
            alternateButton = NSString.stringWith(DWT.getMessage("SWT_Cancel"));
            break;
        case DWT.YES:
            defaultButton = NSString.stringWith(DWT.getMessage("SWT_Yes"));
            break;
        case DWT.NO:
            defaultButton = NSString.stringWith(DWT.getMessage("SWT_No"));
            break;
        case DWT.YES | DWT.NO:
            defaultButton = NSString.stringWith(DWT.getMessage("SWT_Yes"));
            alternateButton = NSString.stringWith(DWT.getMessage("SWT_No"));
            break;
        case DWT.YES | DWT.NO | DWT.CANCEL:             
            defaultButton = NSString.stringWith(DWT.getMessage("SWT_Yes"));
            alternateButton = NSString.stringWith(DWT.getMessage("SWT_No"));
            otherButton = NSString.stringWith(DWT.getMessage("SWT_Cancel"));
            break;
        case DWT.RETRY | DWT.CANCEL:
            defaultButton = NSString.stringWith(DWT.getMessage("SWT_Retry"));
            alternateButton = NSString.stringWith(DWT.getMessage("SWT_Cancel"));
            break;
        case DWT.ABORT | DWT.RETRY | DWT.IGNORE:
            defaultButton = NSString.stringWith(DWT.getMessage("SWT_Abort"));
            alternateButton = NSString.stringWith(DWT.getMessage("SWT_Retry"));
            otherButton = NSString.stringWith(DWT.getMessage("SWT_Ignore"));
            break;
    }
    NSString title = NSString.stringWith(this.title !is null ? this.title : "");
    NSString message = NSString.stringWith(this.message !is null ? this.message : "");
    NSAlert alert = NSAlert.alertWithMessageText(title, defaultButton, alternateButton, otherButton, message);
    alert.setAlertStyle(alertType);
    int response = alert.runModal();
    switch (bits) {
        case DWT.OK:
            switch (response) {
                case OS.NSAlertDefaultReturn:
                    return DWT.OK;
            }
            break;
        case DWT.CANCEL:
            switch (response) {
                case OS.NSAlertDefaultReturn:
                    return DWT.CANCEL;
            }
            break;
        case DWT.OK | DWT.CANCEL:
            switch (response) {
                case OS.NSAlertDefaultReturn:
                    return DWT.OK;
                case OS.NSAlertAlternateReturn:
                    return DWT.CANCEL;
            }
            break;
        case DWT.YES:
            switch (response) {
                case OS.NSAlertDefaultReturn:
                    return DWT.YES;
            }
            break;
        case DWT.NO:
            switch (response) {
                case OS.NSAlertDefaultReturn:
                    return DWT.NO;
            }
            break;
        case DWT.YES | DWT.NO:
            switch (response) {
                case OS.NSAlertDefaultReturn:
                    return DWT.YES;
                case OS.NSAlertAlternateReturn:
                    return DWT.NO;
            }
            break;
        case DWT.YES | DWT.NO | DWT.CANCEL:             
            switch (response) {
                case OS.NSAlertDefaultReturn:
                    return DWT.YES;
                case OS.NSAlertAlternateReturn:
                    return DWT.NO;
                case OS.NSAlertOtherReturn:
                    return DWT.CANCEL;
            }
            break;
        case DWT.RETRY | DWT.CANCEL:
            switch (response) {
                case OS.NSAlertDefaultReturn:
                    return DWT.RETRY;
                case OS.NSAlertAlternateReturn:
                    return DWT.CANCEL;
            }
            break;
        case DWT.ABORT | DWT.RETRY | DWT.IGNORE:
            switch (response) {
                case OS.NSAlertDefaultReturn:
                    return DWT.ABORT;
                case OS.NSAlertAlternateReturn:
                    return DWT.RETRY;
                case OS.NSAlertOtherReturn:
                    return DWT.IGNORE;
            }
            break;
    }
    return DWT.CANCEL;
}

/**
 * Sets the dialog's message, which is a description of
 * the purpose for which it was opened. This message will be
 * visible on the dialog while it is open.
 *
 * @param string the message
 * 
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_NULL_ARGUMENT - if the string is null</li>
 * </ul>
 */
public void setMessage (String string) {
    if (string is null) error (DWT.ERROR_NULL_ARGUMENT);
    message = string;
}

}
