/*******************************************************************************
 * Copyright (c) 2000, 2007 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
module dwt.widgets.Shell;

import dwt.dwthelper.utils;

import dwt.DWT;
import dwt.DWTException;
import dwt.events.ShellListener;
import dwt.graphics.Cursor;
import dwt.graphics.GC;
import dwt.graphics.Point;
import dwt.graphics.Rectangle;
import dwt.graphics.Region;
import dwt.internal.cocoa.NSEvent;
import dwt.internal.cocoa.NSPoint;
import dwt.internal.cocoa.NSRect;
import dwt.internal.cocoa.NSScrollView;
import dwt.internal.cocoa.NSSize;
import dwt.internal.cocoa.NSString;
import dwt.internal.cocoa.NSView;
import dwt.internal.cocoa.NSWindow;
import dwt.internal.cocoa.OS;
import dwt.internal.cocoa.SWTWindow;
import dwt.internal.cocoa.SWTWindowDelegate;

/**
 * Instances of this class represent the "windows"
 * which the desktop or "window manager" is managing.
 * Instances that do not have a parent (that is, they
 * are built using the constructor, which takes a 
 * <code>Display</code> as the argument) are described
 * as <em>top level</em> shells. Instances that do have
 * a parent are described as <em>secondary</em> or
 * <em>dialog</em> shells.
 * <p>
 * Instances are always displayed in one of the maximized, 
 * minimized or normal states:
 * <ul>
 * <li>
 * When an instance is marked as <em>maximized</em>, the
 * window manager will typically resize it to fill the
 * entire visible area of the display, and the instance
 * is usually put in a state where it can not be resized 
 * (even if it has style <code>RESIZE</code>) until it is
 * no longer maximized.
 * </li><li>
 * When an instance is in the <em>normal</em> state (neither
 * maximized or minimized), its appearance is controlled by
 * the style constants which were specified when it was created
 * and the restrictions of the window manager (see below).
 * </li><li>
 * When an instance has been marked as <em>minimized</em>,
 * its contents (client area) will usually not be visible,
 * and depending on the window manager, it may be
 * "iconified" (that is, replaced on the desktop by a small
 * simplified representation of itself), relocated to a
 * distinguished area of the screen, or hidden. Combinations
 * of these changes are also possible.
 * </li>
 * </ul>
 * </p><p>
 * The <em>modality</em> of an instance may be specified using
 * style bits. The modality style bits are used to determine
 * whether input is blocked for other shells on the display.
 * The <code>PRIMARY_MODAL</code> style allows an instance to block
 * input to its parent. The <code>APPLICATION_MODAL</code> style
 * allows an instance to block input to every other shell in the
 * display. The <code>SYSTEM_MODAL</code> style allows an instance
 * to block input to all shells, including shells belonging to
 * different applications.
 * </p><p>
 * Note: The styles supported by this class are treated
 * as <em>HINT</em>s, since the window manager for the
 * desktop on which the instance is visible has ultimate
 * control over the appearance and behavior of decorations
 * and modality. For example, some window managers only
 * support resizable windows and will always assume the
 * RESIZE style, even if it is not set. In addition, if a
 * modality style is not supported, it is "upgraded" to a
 * more restrictive modality style that is supported. For
 * example, if <code>PRIMARY_MODAL</code> is not supported,
 * it would be upgraded to <code>APPLICATION_MODAL</code>.
 * A modality style may also be "downgraded" to a less
 * restrictive style. For example, most operating systems
 * no longer support <code>SYSTEM_MODAL</code> because
 * it can freeze up the desktop, so this is typically
 * downgraded to <code>APPLICATION_MODAL</code>.
 * <dl>
 * <dt><b>Styles:</b></dt>
 * <dd>BORDER, CLOSE, MIN, MAX, NO_TRIM, RESIZE, TITLE, ON_TOP, TOOL</dd>
 * <dd>APPLICATION_MODAL, MODELESS, PRIMARY_MODAL, SYSTEM_MODAL</dd>
 * <dt><b>Events:</b></dt>
 * <dd>Activate, Close, Deactivate, Deiconify, Iconify</dd>
 * </dl>
 * Class <code>DWT</code> provides two "convenience constants"
 * for the most commonly required style combinations:
 * <dl>
 * <dt><code>SHELL_TRIM</code></dt>
 * <dd>
 * the result of combining the constants which are required
 * to produce a typical application top level shell: (that 
 * is, <code>CLOSE | TITLE | MIN | MAX | RESIZE</code>)
 * </dd>
 * <dt><code>DIALOG_TRIM</code></dt>
 * <dd>
 * the result of combining the constants which are required
 * to produce a typical application dialog shell: (that 
 * is, <code>TITLE | CLOSE | BORDER</code>)
 * </dd>
 * </dl>
 * </p>
 * <p>
 * Note: Only one of the styles APPLICATION_MODAL, MODELESS, 
 * PRIMARY_MODAL and SYSTEM_MODAL may be specified.
 * </p><p>
 * IMPORTANT: This class is not intended to be subclassed.
 * </p>
 *
 * @see Decorations
 * @see DWT
 */
public class Shell : Decorations {
    NSWindow window;
    SWTWindowDelegate windowDelegate;
    bool opened, moved, resized, fullScreen;
//  bool resized, moved, drawing, reshape, update, deferDispose, active, disposed, opened, fullScreen;
    Control lastActive;
    Region region;
    Rectangle normalBounds;

    static int DEFAULT_CLIENT_WIDTH = -1;
    static int DEFAULT_CLIENT_HEIGHT = -1;

/**
 * Constructs a new instance of this class. This is equivalent
 * to calling <code>Shell((Display) null)</code>.
 *
 * @exception DWTException <ul>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the parent</li>
 *    <li>ERROR_INVALID_SUBCLASS - if this class is not an allowed subclass</li>
 * </ul>
 */
public this () {
    this ((Display) null);
}

/**
 * Constructs a new instance of this class given only the style
 * value describing its behavior and appearance. This is equivalent
 * to calling <code>Shell((Display) null, style)</code>.
 * <p>
 * The style value is either one of the style constants defined in
 * class <code>DWT</code> which is applicable to instances of this
 * class, or must be built by <em>bitwise OR</em>'ing together 
 * (that is, using the <code>int</code> "|" operator) two or more
 * of those <code>DWT</code> style constants. The class description
 * lists the style constants that are applicable to the class.
 * Style bits are also inherited from superclasses.
 * </p>
 *
 * @param style the style of control to construct
 *
 * @exception DWTException <ul>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the parent</li>
 *    <li>ERROR_INVALID_SUBCLASS - if this class is not an allowed subclass</li>
 * </ul>
 * 
 * @see DWT#BORDER
 * @see DWT#CLOSE
 * @see DWT#MIN
 * @see DWT#MAX
 * @see DWT#RESIZE
 * @see DWT#TITLE
 * @see DWT#NO_TRIM
 * @see DWT#SHELL_TRIM
 * @see DWT#DIALOG_TRIM
 * @see DWT#MODELESS
 * @see DWT#PRIMARY_MODAL
 * @see DWT#APPLICATION_MODAL
 * @see DWT#SYSTEM_MODAL
 */
public this (int style) {
    this ((Display) null, style);
}

/**
 * Constructs a new instance of this class given only the display
 * to create it on. It is created with style <code>DWT.SHELL_TRIM</code>.
 * <p>
 * Note: Currently, null can be passed in for the display argument.
 * This has the effect of creating the shell on the currently active
 * display if there is one. If there is no current display, the 
 * shell is created on a "default" display. <b>Passing in null as
 * the display argument is not considered to be good coding style,
 * and may not be supported in a future release of DWT.</b>
 * </p>
 *
 * @param display the display to create the shell on
 *
 * @exception DWTException <ul>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the parent</li>
 *    <li>ERROR_INVALID_SUBCLASS - if this class is not an allowed subclass</li>
 * </ul>
 */
public this (Display display) {
    this (display, DWT.SHELL_TRIM);
}

/**
 * Constructs a new instance of this class given the display
 * to create it on and a style value describing its behavior
 * and appearance.
 * <p>
 * The style value is either one of the style constants defined in
 * class <code>DWT</code> which is applicable to instances of this
 * class, or must be built by <em>bitwise OR</em>'ing together 
 * (that is, using the <code>int</code> "|" operator) two or more
 * of those <code>DWT</code> style constants. The class description
 * lists the style constants that are applicable to the class.
 * Style bits are also inherited from superclasses.
 * </p><p>
 * Note: Currently, null can be passed in for the display argument.
 * This has the effect of creating the shell on the currently active
 * display if there is one. If there is no current display, the 
 * shell is created on a "default" display. <b>Passing in null as
 * the display argument is not considered to be good coding style,
 * and may not be supported in a future release of DWT.</b>
 * </p>
 *
 * @param display the display to create the shell on
 * @param style the style of control to construct
 *
 * @exception DWTException <ul>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the parent</li>
 *    <li>ERROR_INVALID_SUBCLASS - if this class is not an allowed subclass</li>
 * </ul>
 * 
 * @see DWT#BORDER
 * @see DWT#CLOSE
 * @see DWT#MIN
 * @see DWT#MAX
 * @see DWT#RESIZE
 * @see DWT#TITLE
 * @see DWT#NO_TRIM
 * @see DWT#SHELL_TRIM
 * @see DWT#DIALOG_TRIM
 * @see DWT#MODELESS
 * @see DWT#PRIMARY_MODAL
 * @see DWT#APPLICATION_MODAL
 * @see DWT#SYSTEM_MODAL
 */
public this (Display display, int style) {
    this (display, null, style, 0, false);
}

this (Display display, Shell parent, int style, int handle, bool embedded) {
    super ();
    checkSubclass ();
    if (display is null) display = Display.getCurrent ();
    if (display is null) display = Display.getDefault ();
    if (!display.isValidThread ()) {
        error (DWT.ERROR_THREAD_INVALID_ACCESS);
    }
    if (parent !is null && parent.isDisposed ()) {
        error (DWT.ERROR_INVALID_ARGUMENT); 
    }
    this.style = checkStyle (style);
    this.parent = parent;
    this.display = display;
    if (handle !is 0) {
        if (embedded) {
            view = new NSView(handle);
        } else {
            window = new NSWindow(handle);
            state |= FOREIGN_HANDLE;
        }
    }
    createWidget ();
}

/**
 * Constructs a new instance of this class given only its
 * parent. It is created with style <code>DWT.DIALOG_TRIM</code>.
 * <p>
 * Note: Currently, null can be passed in for the parent.
 * This has the effect of creating the shell on the currently active
 * display if there is one. If there is no current display, the 
 * shell is created on a "default" display. <b>Passing in null as
 * the parent is not considered to be good coding style,
 * and may not be supported in a future release of DWT.</b>
 * </p>
 *
 * @param parent a shell which will be the parent of the new instance
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_INVALID_ARGUMENT - if the parent is disposed</li> 
 * </ul>
 * @exception DWTException <ul>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the parent</li>
 *    <li>ERROR_INVALID_SUBCLASS - if this class is not an allowed subclass</li>
 * </ul>
 */
public this (Shell parent) {
    this (parent, DWT.DIALOG_TRIM);
}

/**
 * Constructs a new instance of this class given its parent
 * and a style value describing its behavior and appearance.
 * <p>
 * The style value is either one of the style constants defined in
 * class <code>DWT</code> which is applicable to instances of this
 * class, or must be built by <em>bitwise OR</em>'ing together 
 * (that is, using the <code>int</code> "|" operator) two or more
 * of those <code>DWT</code> style constants. The class description
 * lists the style constants that are applicable to the class.
 * Style bits are also inherited from superclasses.
 * </p><p>
 * Note: Currently, null can be passed in for the parent.
 * This has the effect of creating the shell on the currently active
 * display if there is one. If there is no current display, the 
 * shell is created on a "default" display. <b>Passing in null as
 * the parent is not considered to be good coding style,
 * and may not be supported in a future release of DWT.</b>
 * </p>
 *
 * @param parent a shell which will be the parent of the new instance
 * @param style the style of control to construct
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_INVALID_ARGUMENT - if the parent is disposed</li> 
 * </ul>
 * @exception DWTException <ul>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the parent</li>
 *    <li>ERROR_INVALID_SUBCLASS - if this class is not an allowed subclass</li>
 * </ul>
 * 
 * @see DWT#BORDER
 * @see DWT#CLOSE
 * @see DWT#MIN
 * @see DWT#MAX
 * @see DWT#RESIZE
 * @see DWT#TITLE
 * @see DWT#NO_TRIM
 * @see DWT#SHELL_TRIM
 * @see DWT#DIALOG_TRIM
 * @see DWT#ON_TOP
 * @see DWT#TOOL
 * @see DWT#MODELESS
 * @see DWT#PRIMARY_MODAL
 * @see DWT#APPLICATION_MODAL
 * @see DWT#SYSTEM_MODAL
 */
public this (Shell parent, int style) {
    this (parent !is null ? parent.display : null, parent, style, 0, false);
}

public static Shell internal_new (Display display, int handle) {
    return new Shell (display, null, DWT.NO_TRIM, handle, false);
}

static int checkStyle (int style) {
    style = Decorations.checkStyle (style);
    style &= ~DWT.TRANSPARENT;
    int mask = DWT.SYSTEM_MODAL | DWT.APPLICATION_MODAL | DWT.PRIMARY_MODAL;
    int bits = style & ~mask;
    if ((style & DWT.SYSTEM_MODAL) !is 0) return bits | DWT.SYSTEM_MODAL;
    if ((style & DWT.APPLICATION_MODAL) !is 0) return bits | DWT.APPLICATION_MODAL;
    if ((style & DWT.PRIMARY_MODAL) !is 0) return bits | DWT.PRIMARY_MODAL;
    return bits;
}

/**
 * Adds the listener to the collection of listeners who will
 * be notified when operations are performed on the receiver,
 * by sending the listener one of the messages defined in the
 * <code>ShellListener</code> interface.
 *
 * @param listener the listener which should be notified
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_NULL_ARGUMENT - if the listener is null</li>
 * </ul>
 * @exception DWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 *
 * @see ShellListener
 * @see #removeShellListener
 */
public void addShellListener(ShellListener listener) {
    checkWidget();
    if (listener is null) error (DWT.ERROR_NULL_ARGUMENT);
    TypedListener typedListener = new TypedListener (listener);
    addListener(DWT.Activate,typedListener);
    addListener(DWT.Close,typedListener);
    addListener(DWT.Deactivate,typedListener);
    addListener(DWT.Iconify,typedListener);
    addListener(DWT.Deiconify,typedListener);
}

void bringToTop (bool force) {
    if (getMinimized ()) return;
    if (force) {
        forceActive ();
    } else {
        setActive ();
    }
}

void checkOpen () {
    if (!opened) resized = false;
}

/**
 * Requests that the window manager close the receiver in
 * the same way it would be closed when the user clicks on
 * the "close box" or performs some other platform specific
 * key or mouse combination that indicates the window
 * should be removed.
 *
 * @exception DWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 *
 * @see DWT#Close
 * @see #dispose
 */
public void close () {
    checkWidget();
    closeWidget ();
}

void closeWidget () {
    Event event = new Event ();
    sendEvent (DWT.Close, event);
    if (event.doit && !isDisposed ()) dispose ();
}

public Rectangle computeTrim (int x, int y, int width, int height) {
    checkWidget();
    Rectangle trim = super.computeTrim(x, y, width, height);
    NSRect rect = new NSRect ();
    rect.x = trim.x;
    rect.y = trim.y;
    rect.width = trim.width;
    rect.height = trim.height;
    rect = NSWindow.static_frameRectForContentRect_styleMask_(rect, window.styleMask());
    return new Rectangle ((int)rect.x, (int)rect.y, (int)rect.width, (int)rect.height);
}

void createHandle () {
    state |= CANVAS;// | GRAB | HIDDEN;
    if (window !is null) {
        view = window.contentView();
        return;
    } else {
        SWTWindow swtWindow = (SWTWindow) new SWTWindow ().alloc ();
        swtWindow.setTag(jniRef);
        window = (NSWindow)swtWindow;
        NSRect rect = new NSRect();
        Monitor monitor = getMonitor ();
        Rectangle clientArea = monitor.getClientArea ();
        rect.width = clientArea.width * 5 / 8;
        rect.height = clientArea.height * 5 / 8;
        int styleMask = OS.NSBorderlessWindowMask;
        if ((style & DWT.NO_TRIM) is 0) {
            styleMask = OS.NSTitledWindowMask;
            if ((style & DWT.CLOSE) !is 0) styleMask |= OS.NSClosableWindowMask;
            if ((style & DWT.MIN) !is 0) styleMask |= OS.NSMiniaturizableWindowMask;
            if ((style & DWT.MAX) !is 0) styleMask |= OS.NSResizableWindowMask;
            if ((style & DWT.RESIZE) !is 0) styleMask |= OS.NSResizableWindowMask;
        }
        window = window.initWithContentRect_styleMask_backing_defer_(rect, styleMask, OS.NSBackingStoreBuffered, false);
        display.cascade = window.cascadeTopLeftFromPoint(display.cascade);
        if ((style & DWT.ON_TOP) !is 0) {
            window.setLevel(OS.NSFloatingWindowLevel);
        }
    }
    
    createHandle (null);
    
    window.setContentView (topView());
    windowDelegate = (SWTWindowDelegate)new SWTWindowDelegate().alloc().init();
    windowDelegate.setTag(jniRef);
    window.setDelegate(windowDelegate);
}

void destroyWidget () {
    NSWindow window = this.window;
    releaseHandle ();
    if (window !is null) window.close();
}

Control findBackgroundControl () {
    return background !is null || backgroundImage !is null ? this : null;
}

Composite findDeferredControl () {
    return layoutCount > 0 ? this : null;
}

Cursor findCursor () {
    return cursor;
}

void fixShell (Shell newShell, Control control) {
    if (this is newShell) return;
//  if (control is lastActive) setActiveControl (null);
}

void flagsChanged(int theEvent) {
    Display display = this.display;
    NSEvent nsEvent = new NSEvent(theEvent);
    int modifiers = nsEvent.modifierFlags();
    int lastModifiers = display.lastModifiers;
//  int chord = OS.GetCurrentEventButtonState ();
    int type = DWT.KeyUp;   
    if ((modifiers & OS.NSAlphaShiftKeyMask) !is 0 && (lastModifiers & OS.NSAlphaShiftKeyMask) is 0) type = DWT.KeyDown;
    if ((modifiers & OS.NSAlternateKeyMask) !is 0 && (lastModifiers & OS.NSAlternateKeyMask) is 0) type = DWT.KeyDown;
    if ((modifiers & OS.NSShiftKeyMask) !is 0 && (lastModifiers & OS.NSShiftKeyMask) is 0) type = DWT.KeyDown;
    if ((modifiers & OS.NSControlKeyMask) !is 0 && (lastModifiers & OS.NSControlKeyMask) is 0) type = DWT.KeyDown;
    if ((modifiers & OS.NSCommandKeyMask) !is 0 && (lastModifiers & OS.NSCommandKeyMask) is 0) type = DWT.KeyDown;
    Control target = display.getFocusControl();
    if (type is DWT.KeyUp && (modifiers & OS.NSAlphaShiftKeyMask) is 0 && (lastModifiers & OS.NSAlphaShiftKeyMask) !is 0) {
        if (target !is null) {
            Event event = new Event ();
            event.keyCode = DWT.CAPS_LOCK;
    //      setInputState (event, DWT.KeyDown, chord, modifiers);
            target.sendKeyEvent (DWT.KeyDown, event);
        }
    }
    Event event = new Event ();
//  setInputState (event, type, chord, modifiers);
    if (event.keyCode is 0 && event.character is 0) return;
    bool result = sendKeyEvent (type, event);
    if (type is DWT.KeyDown && (modifiers & OS.NSAlphaShiftKeyMask) !is 0 && (lastModifiers & OS.NSAlphaShiftKeyMask) is 0) {
        if (target !is null) {
            event = new Event ();
            event.keyCode = DWT.CAPS_LOCK;
    //      setInputState (event, DWT.KeyUp, chord, modifiers);
            target.sendKeyEvent (DWT.KeyUp, event);
        }
    }
    display.lastModifiers = modifiers;
}

/**
 * If the receiver is visible, moves it to the top of the 
 * drawing order for the display on which it was created 
 * (so that all other shells on that display, which are not 
 * the receiver's children will be drawn behind it) and forces 
 * the window manager to make the shell active.
 *
 * @exception DWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 *
 * @since 2.0
 * @see Control#moveAbove
 * @see Control#setFocus
 * @see Control#setVisible
 * @see Display#getActiveShell
 * @see Decorations#setDefaultButton(Button)
 * @see Shell#open
 * @see Shell#setActive
 */
public void forceActive () {
    checkWidget ();
    if (!isVisible ()) return;
//  OS.SelectWindow (shellHandle);
//  OS.SetFrontProcessWithOptions (new int [] {0, OS.kCurrentProcess}, OS.kSetFrontProcessFrontWindowOnly);
}

public int getAlpha () {
    checkWidget ();
    return (int)(window.alphaValue() * 255);
}

public Rectangle getBounds () {
    checkWidget();
    NSRect frame = window.frame ();
    return new Rectangle ((int)frame.x, (int) frame.y, (int) frame.width, (int) frame.height);
}

public Rectangle getClientArea () {
    checkWidget();
    //TODO why super implementation fails
    NSRect rect = window.contentRectForFrameRect_(window.frame());
    int width = (int)rect.width, height = (int)rect.height;
    if (scrollView !is null) {
        NSSize size = new NSSize();
        size.width = width;
        size.height = height;
        size = NSScrollView.contentSizeForFrameSize(size, (style & DWT.H_SCROLL) !is 0, (style & DWT.V_SCROLL) !is 0, OS.NSNoBorder);
        width = (int)size.width;
        height = (int)size.height;
    }
    return new Rectangle (0, 0, width, height);
}

int getDrawCount (int control) {
    if (!isTrimHandle (control)) return drawCount;
    return 0;
}

public bool getFullScreen () {
    checkWidget();
    return fullScreen;
}

/**
 * Returns the receiver's input method editor mode. This
 * will be the result of bitwise OR'ing together one or
 * more of the following constants defined in class
 * <code>DWT</code>:
 * <code>NONE</code>, <code>ROMAN</code>, <code>DBCS</code>, 
 * <code>PHONETIC</code>, <code>NATIVE</code>, <code>ALPHA</code>.
 *
 * @return the IME mode
 *
 * @exception DWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 *
 * @see DWT
 */
public int getImeInputMode () {
    checkWidget();
    return DWT.NONE;
}

public Point getLocation () {
    checkWidget();
    NSRect frame = window.frame ();
    return new Point ((int) frame.x, (int) frame.y);
}

public bool getMaximized () {
    checkWidget();
    //NOT DONE
    return !fullScreen && super.getMaximized ();
}

public bool getMinimized () {
    checkWidget();
    if (!getVisible ()) return super.getMinimized ();
//  return OS.IsWindowCollapsed (shellHandle);
    return false;
}

/**
 * Returns a point describing the minimum receiver's size. The
 * x coordinate of the result is the minimum width of the receiver.
 * The y coordinate of the result is the minimum height of the
 * receiver.
 *
 * @return the receiver's size
 *
 * @exception DWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 * 
 * @since 3.1
 */
public Point getMinimumSize () {
    checkWidget();
//  Rect rect = new Rect ();
//  OS.GetWindowStructureWidths (shellHandle, rect);
//  CGPoint inMinLimits = new CGPoint (), inMaxLimits = new CGPoint ();
//  OS.GetWindowResizeLimits (shellHandle, inMinLimits, inMaxLimits);
//  int width = Math.max (1, (int) inMinLimits.x + (rect.left + rect.right));
//  int height = Math.max (1, (int) inMinLimits.y + (rect.top + rect.bottom));
//  return new Point (width, height);
    return null;
}

float [] getParentBackground () {
    return null;
}

/** 
 * Returns the region that defines the shape of the shell,
 * or null if the shell has the default shape.
 *
 * @return the region that defines the shape of the shell (or null)
 *  
 * @exception DWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 *
 * @since 3.0
 *
 */
public Region getRegion () {
    /* This method is needed for the @since 3.0 Javadoc */
    checkWidget ();
    return region;
}

public Shell getShell () {
    checkWidget();
    return this;
}

/**
 * Returns an array containing all shells which are 
 * descendants of the receiver.
 * <p>
 * @return the dialog shells
 *
 * @exception DWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 */
public Shell [] getShells () {
    checkWidget();
    int count = 0;
    Shell [] shells = display.getShells ();
    for (int i=0; i<shells.length; i++) {
        Control shell = shells [i];
        do {
            shell = shell.parent;
        } while (shell !is null && shell !is this);
        if (shell is this) count++;
    }
    int index = 0;
    Shell [] result = new Shell [count];
    for (int i=0; i<shells.length; i++) {
        Control shell = shells [i];
        do {
            shell = shell.parent;
        } while (shell !is null && shell !is this);
        if (shell is this) {
            result [index++] = shells [i];
        }
    }
    return result;
}

public Point getSize () {
    checkWidget();
    NSRect frame = window.frame ();
    return new Point ((int) frame.width, (int) frame.height);
}

bool hasBorder () {
    return false;
}

void helpRequested(int theEvent) {
    Control control = display.getFocusControl();
    while (control !is null) {
        if (control.hooks (DWT.Help)) {
            control.postEvent (DWT.Help);
            break;
        }
        control = control.parent;
    }
}

public bool isEnabled () {
    checkWidget();
    return getEnabled ();
}

bool isEnabledCursor () {
    return true;
}

public bool isVisible () {
    checkWidget();
    return getVisible ();
}

/**
 * Moves the receiver to the top of the drawing order for
 * the display on which it was created (so that all other
 * shells on that display, which are not the receiver's
 * children will be drawn behind it), marks it visible,
 * sets the focus and asks the window manager to make the
 * shell active.
 *
 * @exception DWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 *
 * @see Control#moveAbove
 * @see Control#setFocus
 * @see Control#setVisible
 * @see Display#getActiveShell
 * @see Decorations#setDefaultButton(Button)
 * @see Shell#setActive
 * @see Shell#forceActive
 */
public void open () {
    checkWidget();
    setWindowVisible (true, true);
    if (isDisposed ()) return;
//  if (active) {
        if (!restoreFocus () && !traverseGroup (true)) setFocus ();
//  }
}

public bool print (GC gc) {
    checkWidget ();
    if (gc is null) error (DWT.ERROR_NULL_ARGUMENT);
    if (gc.isDisposed ()) error (DWT.ERROR_INVALID_ARGUMENT);
//  int [] outImage = new int [1];
//  CGRect outFrame = new CGRect ();
//  if (OS.HIViewCreateOffscreenImage (handle, 0, outFrame, outImage) is OS.noErr) {
//      int width = OS.CGImageGetWidth (outImage [0]);
//      int height = OS.CGImageGetHeight (outImage [0]);
//      CGRect rect = new CGRect();
//      rect.width = width;
//      rect.height = height;
//      //TODO - does not draw the browser (cocoa widgets?)
//      OS.HIViewDrawCGImage (gc.handle, rect, outImage [0]);
//      OS.CGImageRelease (outImage [0]);
//  }
//  return true;
    return false;
}

void releaseChildren (bool destroy) {
    Shell [] shells = getShells ();
    for (int i=0; i<shells.length; i++) {
        Shell shell = shells [i];
        if (shell !is null && !shell.isDisposed ()) {
            shell.dispose ();
        }
    }
    super.releaseChildren (destroy);
}

void releaseHandle () {
    window.setDelegate(null);
    if (windowDelegate !is null) windowDelegate.release();
    windowDelegate = null;
    super.releaseHandle ();
    window = null;
}

void releaseParent () {
    /* Do nothing */
}

void releaseWidget () {
    super.releaseWidget ();
//  disposed = true;
    lastActive = null;
}

/**
 * Removes the listener from the collection of listeners who will
 * be notified when operations are performed on the receiver.
 *
 * @param listener the listener which should no longer be notified
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_NULL_ARGUMENT - if the listener is null</li>
 * </ul>
 * @exception DWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 *
 * @see ShellListener
 * @see #addShellListener
 */
public void removeShellListener(ShellListener listener) {
    checkWidget();
    if (listener is null) error (DWT.ERROR_NULL_ARGUMENT);
    if (eventTable is null) return;
    eventTable.unhook(DWT.Activate, listener);
    eventTable.unhook(DWT.Close, listener);
    eventTable.unhook(DWT.Deactivate, listener);
    eventTable.unhook(DWT.Iconify,listener);
    eventTable.unhook(DWT.Deiconify,listener);
}

/**
 * If the receiver is visible, moves it to the top of the 
 * drawing order for the display on which it was created 
 * (so that all other shells on that display, which are not 
 * the receiver's children will be drawn behind it) and asks 
 * the window manager to make the shell active 
 *
 * @exception DWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 *
 * @since 2.0
 * @see Control#moveAbove
 * @see Control#setFocus
 * @see Control#setVisible
 * @see Display#getActiveShell
 * @see Decorations#setDefaultButton(Button)
 * @see Shell#open
 * @see Shell#setActive
 */
public void setActive () {
    checkWidget ();
    if (!isVisible ()) return;
//  OS.SelectWindow (shellHandle);
}

void setActiveControl (Control control) {
    if (control !is null && control.isDisposed ()) control = null;
    if (lastActive !is null && lastActive.isDisposed ()) lastActive = null;
    if (lastActive is control) return;
    
    /*
    * Compute the list of controls to be activated and
    * deactivated by finding the first common parent
    * control.
    */
    Control [] activate = (control is null) ? new Control[0] : control.getPath ();
    Control [] deactivate = (lastActive is null) ? new Control[0] : lastActive.getPath ();
    lastActive = control;
    int index = 0, length = Math.min (activate.length, deactivate.length);
    while (index < length) {
        if (activate [index] !is deactivate [index]) break;
        index++;
    }
    
    /*
    * It is possible (but unlikely), that application
    * code could have destroyed some of the widgets. If
    * this happens, keep processing those widgets that
    * are not disposed.
    */
    for (int i=deactivate.length-1; i>=index; --i) {
        if (!deactivate [i].isDisposed ()) {
            deactivate [i].sendEvent (DWT.Deactivate);
        }
    }
    for (int i=activate.length-1; i>=index; --i) {
        if (!activate [i].isDisposed ()) {
            activate [i].sendEvent (DWT.Activate);
        }
    }
}

public void setAlpha (int alpha) {
    checkWidget ();
    alpha &= 0xFF;
    window.setAlphaValue (alpha / 255f);
}

int setBounds (int x, int y, int width, int height, bool move, bool resize) {
//  if (fullScreen) setFullScreen (false);
    if (move && resize) {
        NSRect rect = new NSRect ();
        rect.x = x;
        //TODO - get the screen for the point
        int screenHeight = (int) window.screen().frame().height;
        rect.y = screenHeight - y;
        rect.width = width;
        rect.height = height;
        window.setFrame_display_(rect, false);
    } else {
        if (move) {
            NSPoint point = new NSPoint();
            point.x = x;
            //TODO - get the screen for the point
            int screenHeight = (int) window.screen().frame().height;
            point.y = screenHeight - y;
            window.setFrameTopLeftPoint (point);
        } else {
            if (resize) {
                NSRect rect = window.frame();
                rect.width = width;
                rect.height = height;
                window.setFrame_display_(rect, false);
            }
        }
    }
    return 0;
}

public void setEnabled (bool enabled) {
    checkWidget();
    if (((state & DISABLED) is 0) is enabled) return;
    super.setEnabled (enabled);
//  if (enabled && OS.IsWindowActive (shellHandle)) {
//      if (!restoreFocus ()) traverseGroup (false);
//  }
}

public void setFullScreen (bool fullScreen) {
    checkWidget ();
    this.fullScreen = fullScreen; 
//  if (fullScreen) {
//      normalBounds = getBounds ();
//      OS.ChangeWindowAttributes (shellHandle, OS.kWindowNoTitleBarAttribute, OS.kWindowResizableAttribute | OS.kWindowLiveResizeAttribute);
//      updateSystemUIMode ();
//      Rectangle screen = getMonitor ().getBounds ();
//      if (menuBar !is null && getMonitor ().equals(display.getPrimaryMonitor ())) {
//          Rect rect = new Rect ();
//          int gdevice = OS.GetMainDevice ();
//          OS.GetAvailableWindowPositioningBounds (gdevice, rect);
//          screen.height -= rect.top;
//          screen.y += rect.top;
//      }
//      Rect rect = new Rect ();
//      OS.SetRect (rect, (short) screen.x, (short) screen.y, (short) (screen.x + screen.width), (short) (screen.y + screen.height));
//      OS.SetWindowBounds (shellHandle, (short) OS.kWindowStructureRgn, rect);
//  } else {
//      int attributes = 0;
//      if ((style & DWT.RESIZE) !is 0) {
//          attributes |= OS.kWindowResizableAttribute;
//          /*
//          * Bug in the Macintosh.  For some reason, a window has no title bar
//          * and the kWindowResizableAttribute, no rubber banding feedback is
//          * given while the window is resizing.  The fix is to create the window 
//          * with kWindowLiveResizeAttribute in this case.  This is inconsistent
//          * with other windows, but the user will get feedback when resizing.
//          */
//          if ((style & DWT.TITLE) is 0) attributes |= OS.kWindowLiveResizeAttribute;
//          if (!OS.__BIG_ENDIAN__()) attributes |= OS.kWindowLiveResizeAttribute;
//      }
//      OS.ChangeWindowAttributes (shellHandle, attributes, OS.kWindowNoTitleBarAttribute);
//      OS.SetSystemUIMode (OS.kUIModeNormal, 0);
//      if (maximized) {
//          setMaximized (true);
//      } else {
//          Rect rect = new Rect ();
//          if (normalBounds !is null) OS.SetRect (rect, (short) normalBounds.x, (short) normalBounds.y, (short) (normalBounds.x + normalBounds.width), (short) (normalBounds.y + normalBounds.height));
//          OS.SetWindowBounds (shellHandle, (short) OS.kWindowStructureRgn, rect);
//      }
//      normalBounds = null;
//  }
}

public void setMenuBar (Menu menu) {
    checkWidget();
    super.setMenuBar (menu);
    if (display.getActiveShell () is this) {
        display.setMenuBar (menuBar);
    }
}

/**
 * Sets the input method editor mode to the argument which 
 * should be the result of bitwise OR'ing together one or more
 * of the following constants defined in class <code>DWT</code>:
 * <code>NONE</code>, <code>ROMAN</code>, <code>DBCS</code>, 
 * <code>PHONETIC</code>, <code>NATIVE</code>, <code>ALPHA</code>.
 *
 * @param mode the new IME mode
 *
 * @exception DWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 *
 * @see DWT
 */
public void setImeInputMode (int mode) {
    checkWidget();
}

public void setMaximized (bool maximized) {
    checkWidget();
    super.setMaximized (maximized);
//  dwt.internal.carbon.Point pt = new dwt.internal.carbon.Point ();
//  if (maximized) {
//      Rect rect = new Rect ();
//      int gdevice = OS.GetMainDevice ();
//      OS.GetAvailableWindowPositioningBounds (gdevice, rect);
//      pt.h = (short) (rect.right - rect.left);
//      pt.v = (short) (rect.bottom - rect.top);
//  }
//  short inPartCode = (short) (maximized ? OS.inZoomOut : OS.inZoomIn);
//  OS.ZoomWindowIdeal (shellHandle, inPartCode, pt);
}

public void setMinimized (bool minimized) {
    checkWidget();
    if (this.minimized is minimized) return;
    super.setMinimized (minimized);
//  if (!minimized && OS.IsWindowCollapsed (shellHandle)) {
//      OS.SelectWindow (shellHandle);
//  }
//  OS.CollapseWindow (shellHandle, minimized);
}

/**
 * Sets the receiver's minimum size to the size specified by the arguments.
 * If the new minimum size is larger than the current size of the receiver,
 * the receiver is resized to the new minimum size.
 *
 * @param width the new minimum width for the receiver
 * @param height the new minimum height for the receiver
 *
 * @exception DWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 * 
 * @since 3.1
 */
public void setMinimumSize (int width, int height) {
    checkWidget();
//  int clientWidth = 0, clientHeight = 0;
//  int trim = DWT.TITLE | DWT.CLOSE | DWT.MIN | DWT.MAX;
//  if ((style & DWT.NO_TRIM) is 0 && (style & trim) !is 0) {
//      clientWidth = DEFAULT_CLIENT_WIDTH;
//      clientHeight = DEFAULT_CLIENT_HEIGHT;
//  }
//  Rect rect = new Rect ();
//  OS.GetWindowStructureWidths (shellHandle, rect);
//  CGPoint inMinLimits = new CGPoint (), inMaxLimits = new CGPoint ();
//  OS.GetWindowResizeLimits (shellHandle, inMinLimits, inMaxLimits);
//  width = Math.max (width, clientWidth + rect.left + rect.right);
//  height = Math.max (height, clientHeight + rect.top + rect.bottom);
//  inMinLimits.x = width - (rect.left + rect.right);
//  inMinLimits.y = height - (rect.top + rect.bottom);
//  OS.SetWindowResizeLimits (shellHandle, inMinLimits, inMaxLimits);
//  Point size = getSize ();
//  int newWidth = Math.max (size.x, width), newHeight = Math.max (size.y, height);
//  if (newWidth !is size.x || newHeight !is size.y) setSize (newWidth, newHeight);
}

/**
 * Sets the receiver's minimum size to the size specified by the argument.
 * If the new minimum size is larger than the current size of the receiver,
 * the receiver is resized to the new minimum size.
 *
 * @param size the new minimum size for the receiver
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_NULL_ARGUMENT - if the point is null</li>
 * </ul>
 * @exception DWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 * 
 * @since 3.1
 */
public void setMinimumSize (Point size) {
    checkWidget();
    if (size is null) error (DWT.ERROR_NULL_ARGUMENT);
    setMinimumSize (size.x, size.y);
}

/**
 * Sets the shape of the shell to the region specified
 * by the argument.  When the argument is null, the
 * default shape of the shell is restored.  The shell
 * must be created with the style DWT.NO_TRIM in order
 * to specify a region.
 *
 * @param region the region that defines the shape of the shell (or null)
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_INVALID_ARGUMENT - if the region has been disposed</li>
 * </ul>  
 * @exception DWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 *
 * @since 3.0
 *
 */
public void setRegion (Region region) {
    checkWidget ();
    if ((style & DWT.NO_TRIM) is 0) return;
    if (region !is null && region.isDisposed()) error (DWT.ERROR_INVALID_ARGUMENT);
//  if (region is null) {
//      rgnRect = null;
//  } else {
//      if (rgnRect is null) {
//          rgnRect = new Rect ();
//          OS.GetWindowBounds (shellHandle, (short) OS.kWindowStructureRgn, rgnRect);
//          OS.SetRect (rgnRect, (short) 0, (short) 0, (short) (rgnRect.right - rgnRect.left), (short) (rgnRect.bottom - rgnRect.top)); 
//      }
//  }
//  this.region = region;
//  /*
//  * Bug in the Macintosh.  Calling ReshapeCustomWindow() from a
//  * kEventWindowDrawContent handler originating from ShowWindow()
//  * will deadlock.  The fix is to detected this case and only call
//  * ReshapeCustomWindow() after the default handler is done.
//  */
//  if (drawing) {
//      reshape = true;
//  } else {
//      OS.ReshapeCustomWindow (shellHandle);
//      redrawWidget (handle, true);
//  }
}

public void setText (String string) {
    checkWidget();
    if (string is null) error (DWT.ERROR_NULL_ARGUMENT);
    super.setText (string);
    NSString str = NSString.stringWith(string);
    window.setTitle(str);
//  str.release();
}

public void setVisible (bool visible) {
    checkWidget();
    setWindowVisible (visible, false);
}

void setWindowVisible (bool visible, bool key) {
    if (window.isVisible() is visible) return;
    if (visible) {
        sendEvent (DWT.Show);
        if (isDisposed ()) return;
        if (key) {
            window.makeKeyAndOrderFront (null);
        } else {
            window.orderFront (null);
        }
        opened = true;
        if (!moved) {
            moved = true;
            sendEvent (DWT.Move);
            if (isDisposed ()) return;
        }
        if (!resized) {
            resized = true;
            sendEvent (DWT.Resize);
            if (isDisposed ()) return;
            if (layout !is null) {
                markLayout (false, false);
                updateLayout (false);
            }
        }
    } else {
        window.orderOut (null);
        sendEvent (DWT.Hide);
    }
}

void setZOrder () {
//  if (scrolledHandle !is 0) OS.HIViewAddSubview (scrolledHandle, handle);
}

void setZOrder (Control control, bool above) {
//  if (above) {
//      //NOT DONE - move one window above another
//      OS.BringToFront (shellHandle);
//   } else {
//      int window = control is null ? 0 : OS.GetControlOwner (control.handle);
//      OS.SendBehind (shellHandle, window);
//  }
}

bool traverseEscape () {
    if (parent is null) return false;
    if (!isVisible () || !isEnabled ()) return false;
    close ();
    return true;
}

void updateSystemUIMode () {
    if (!getMonitor ().equals (display.getPrimaryMonitor ())) return;
    bool isActive = false;
    Shell activeShell = display.getActiveShell ();
    Shell current = this;
    while (current !is null) {
        if (current.equals (activeShell)) {
            isActive = true;
            break;
        }
        current = (Shell) current.parent;
    }
    if (!isActive) return;
//  if (fullScreen) {
//      int mode = OS.kUIModeAllHidden;
//      if (menuBar !is null) {
//          mode = OS.kUIModeContentHidden;
//      }
//      OS.SetSystemUIMode (mode, 0);
//  } else {
//      OS.SetSystemUIMode (OS.kUIModeNormal, 0);
//  }
}

void windowDidBecomeKey(int notification) {
    super.windowDidBecomeKey(notification);
    Display display = this.display;
    display.setMenuBar (menuBar);
    sendEvent (DWT.Activate);
//  if (!isDisposed ()) {
//      if (!restoreFocus () && !traverseGroup (true)) setFocus ();
//  }
}

void windowDidMove(int notification) {
    moved = true;
    sendEvent(DWT.Move);
}

void windowDidResize(int notification) {
    resized = true;
    sendEvent (DWT.Resize);
    if (isDisposed ()) return;
    if (layout !is null) {
        markLayout (false, false);
        updateLayout (false);
    }
}

void windowDidResignKey(int notification) {
    super.windowDidResignKey(notification);
    Display display = this.display;
    sendEvent (DWT.Deactivate);
    if (isDisposed ()) return;
//  saveFocus ();
//  if (savedFocus !is null) {
//      /*
//      * Bug in the Macintosh.  When ClearKeyboardFocus() is called,
//      * the control that has focus gets two kEventControlSetFocus
//      * events indicating that focus was lost.  The fix is to ignore
//      * both of these and send the focus lost event explicitly.
//      */
//      display.ignoreFocus = true;
//      OS.ClearKeyboardFocus (shellHandle);
//      display.ignoreFocus = false;
//      if (!savedFocus.isDisposed ()) savedFocus.sendFocusEvent (DWT.FocusOut, false);
//  }
    display.setMenuBar (null);
}

bool windowShouldClose(int window) {
    closeWidget ();
    return false;
}

void windowWillClose(int notification) {
}

void windowSendEvent(int id, int event) {
    NSEvent nsEvent = new NSEvent(event);
    int type = nsEvent.type();
    if (type is OS.NSKeyDown || type is OS.NSKeyUp) {
        Control eventTarget = display.getFocusControl();
        if (eventTarget !is null) {
            if (type is OS.NSKeyDown) {
                bool[] consume = new bool[1];
                int key = nsEvent.keyCode();
                if (eventTarget.translateTraversal(key, nsEvent, consume)) return;
                if (consume[0]) return;
                if (eventTarget.isDisposed()) return;
            }
            if (!eventTarget.sendKeyEvent(nsEvent, type is OS.NSKeyDown ? DWT.KeyDown : DWT.KeyUp)) return;
        }
    }
    super.windowSendEvent(id, event);
}

}
